#ifndef _OACLIENT_CPP_
#define _OACLIENT_CPP_

#include "VEProver.h"
#include "OA.h"
#include "OAClient.h"

//----------------------------------------------------------------------------//
// Constructors
OAClient::OAClient(const VEPublicKey* pk, const VEPublicKey* regularpk,
					const int stat, const cipher_t &signAlg)
	: stat(stat), counter(0), verifiablePK(pk), regularPK(regularpk),
	contract(NULL), warrantyVEMessage(NULL), metadata(NULL),
	signKey(NULL), serversSignPK(NULL), receipt(NULL)
{
	signKey = Signature::Key::generateKey(signAlg);
}

// Copy Constructor
OAClient::OAClient(const OAClient &o)
	: stat(o.stat), counter(o.counter), verifiablePK(o.verifiablePK),
	  regularPK(o.regularPK), id(o.id), coin(o.coin), contract(new OAContract(*o.contract)),
	  warrantyVEMessage(new OAEscrowMessage(*o.warrantyVEMessage)), metadata(new OAMetadata(*o.metadata)),
	  signKey(new Signature::Key(*o.signKey)), serversSignPK(new Signature::Key(*o.serversSignPK)),
	  receipt(new string(*o.receipt)), endorsement(o.endorsement) {}

//----------------------------------------------------------------------------//
// Destructor
OAClient::~OAClient() {
	reset();
	delete signKey;
}

void OAClient::reset() {
	if (contract!=NULL) delete contract;
	if (serversSignPK!=NULL) delete serversSignPK;
	if (warrantyVEMessage!=NULL) delete warrantyVEMessage;
	if (metadata!=NULL) delete metadata;
	if (receipt!=NULL) delete receipt;
	counter=0;
	return;
}

//----------------------------------------------------------------------------//
// Setup
Signature::Key* OAClient::sendSignPK() const {
	return signKey->getPublicKey();
}

string* OAClient::sendFile(const OAContract* contract) {
	// check the contract
	if ((contract->getDue() + OA_CONTRACT_TIME_ELASTICITY < time(NULL) + OA_CONTRACT_PERIOD)
		|| (contract->getMaxUpdates() != OA_MAX_UPDATES))
			return false;
	
	// store contract information
	this->contract = new OAContract(*contract);
	this->id = contract->getID();

	// set the server's signPK from the contract
	serversSignPK = new Signature::Key(*contract->getServersSignPK());
	
	// send file
	string* file = new string;
	file->resize(OA_FILE_SIZE);
	return file;
}

//----------------------------------------------------------------------------//
// Send Verifiable Escrow
OAEscrowMessage* OAClient::sendPaymentEscrow(Wallet *wallet, const ZZ& R, OAMetadata &mm) {
	// check the signature on the proof
	if (!Signature::verify(*serversSignPK, *mm.getSignature(), saveString(counter)+(*mm.getProof()), regularPK->hashAlg))
		return NULL;

	// store metadata message
	this->metadata = &mm;
	
	// create a coin
	makeCoin(wallet, R);
	
	return sendPaymentEscrow();
}

OAEscrowMessage* OAClient::sendPaymentEscrow() {
	// set up prover for forming escrow
	VEProver prover(verifiablePK);
	
	ZZ eCom = coin.getEndorsementCom();
	// the label needs to be the public key for the signature scheme
	// create the verifiable escrow
	VECiphertext* escrow = new VECiphertext(prover.verifiableEncrypt(eCom, 
											endorsement, coin.getCashGroup(),
											saveString(contract),
											verifiablePK->hashAlg, stat));

	return new OAEscrowMessage(coin, escrow, *signKey);
}

void OAClient::makeCoin(Wallet* wallet, const ZZ& R) {
	// get a coin
	Coin coin = wallet->nextCoin(R);
	setCoin(coin);
}

void OAClient::setCoin(const Coin& c) {	
	// separate coin' from endorsement
	coin = c;
	endorsement = coin.getEndorsement();
	coin.unendorse();
}

//----------------------------------------------------------------------------//
// Get Warranty & Send Endorsement
string* OAClient::signCtr(const ZZ &R, OAEscrowMessage* warrantyVEMessage) {
	// check escrow
	if (!warrantyVEMessage->check(verifiablePK, stat, R))
		return NULL;
	// store escrow
	this->warrantyVEMessage = warrantyVEMessage;
	
	return new string (Signature::sign(*signKey, saveString(counter), regularPK->hashAlg));
}

vector<ZZ>* OAClient::sendEndorsement(string* serverSignCtr, string& receipt) {
	// check signature on the counter+metadata
	if (!Signature::verify(*serversSignPK, *serverSignCtr, saveString(counter)+(*metadata->getMetadata()), regularPK->hashAlg))
		return NULL;
	// store in metadata message;
	metadata->setSignature(*serverSignCtr);
	delete serverSignCtr;

	// check receipt (signature on the contract)
	if (!Signature::verify(*serversSignPK, receipt, saveString(*contract), regularPK->hashAlg))
		return NULL;
	
	// store receipt
	this->receipt = &receipt;
	
	return &endorsement;
}

//----------------------------------------------------------------------------//
// Update
OAUpdate* OAClient::update() {
	// create update
	string *update = new string();
	update->resize(OA_UPDATE_SIZE);
	
	// sign counter
	string *sign = new string(Signature::sign(*signKey, saveString(counter+1), regularPK->hashAlg));
	
	// return update message
	return new OAUpdate(update, sign);
}

bool OAClient::checkMetadata(OAMetadata &mm) {
	// check signature on the proof
	if (!Signature::verify(*serversSignPK, *mm.getSignature(), saveString(counter+1)+(*mm.getMetadata()), regularPK->hashAlg))
		return false;
	
	// store metadata message
	if (metadata != NULL)
		delete metadata;
	this->metadata = &mm;

	// increment counter
	counter++;
	
	return true;
}

//----------------------------------------------------------------------------//
// Resolve
OAClaim* OAClient::sendClaim() const {
	// claim only wraps pointers
	return new OAClaim(contract, counter, warrantyVEMessage, receipt, metadata);
}

bool OAClient::checkEndorsement(vector<ZZ> &endorsement) {
	// check endorsement
	bool endorsementOK = warrantyVEMessage->getCoinPrime().endorse(endorsement);
	delete &endorsement;
	return endorsementOK;
}

#endif

