#ifndef _OACLIENT_H_
#define _OACLIENT_H_

#include "Wallet.h"
#include "Ciphertext.h"
#include "OAContract.h"
#include "OAEscrowMessage.h"
#include "OAMetadata.h"
#include "OAClaim.h"
#include "OAUpdate.h"

//! This class runs the client side of the Official Arbitration protocol

class OAClient {
	public:
		//! Constructor
		/**
		 * @param pk 		verifiable public key of arbiter
		 * @param regularpk regular public key of arbiter
		 * @param stat 		statistical security parameter
		 * @param signAlg 	algorithm to be used for signature key generation
		 */
		OAClient(const VEPublicKey* pk, const VEPublicKey* regularpk,
				 const int stat, const cipher_t &signAlg);

		//! Copy Constructor
		OAClient(const OAClient &o);

		//! Destructor
		~OAClient();
		
		//! Reset data members to initilization values
		void reset();
		
		//**********************************************//
		
		// Setup functions
		/*! Send client's public key to server at setup phase
		 * @return client's public key to be used in signature varification
		 */
		Signature::Key* sendSignPK() const;
		/*! Receive contract from server,
		 * if it accepts the contract sends file to be stored
		 * @param contract contract sent by server
		 * @return file to be stored (NULL if contract is not accepted)
		 */
		string* sendFile(const OAContract* contract);

		// Payment functions
		/*! Receive metadata message, and return verifiable 
		 * escrow of payment if metadata message is accepted
		 * @param wallet 			wallet which coin is generated from
		 * @param R 				randomization variable
		 * @param metadataMessage 	metadata message consisting of metadata, proof and server's signature on proof
		 * @return escrow message consisting of unendorsed coin, verifiable escrow and public key of the client (returns NULL if metadata message is not valid)
		 */
		OAEscrowMessage* sendPaymentEscrow(Wallet *wallet, const ZZ& R, OAMetadata &metadataMessage);
		/*! Receive verifiable escrow of warranty and 
		 * return client's signature on counter which is initiated to 0
		 * @param R 				randomization variable
		 * @param warrantyVEMessage	escrow message consisting of unendorsed coin, verifiable escrow and public key of the client
		 * @return client's signature on its own counter (NULL if escrow message is not valid)
		 */
		string* signCtr(const ZZ &R, OAEscrowMessage* warrantyVEMessage);
		/*! Receive receipt and server's signature on counter which is expected to be 0,  
		 * and return the endorsement of payment to complete the payment if both signatures verify under server's public key.
		 * @param serverSignCtr	server's signature on server's counter
		 * @param receipt 		server's signature on contract
		 * @return endorsement of payment coin (NULL if signatures do not verify)
		 */
		vector<ZZ>* sendEndorsement(string* serverSignCtr, string& receipt);
		
		// Update rounds
		/*! After completion of setup phase, send updates for client's file on server 
		 * @return update information
		 */
		OAUpdate* update();
		/*! Receive a new metadata message for its update,
		 * check metadata, proof, and server's signature on metadata
		 * return whether the metadata message is accepted
		 * @return true if metadata message is correctly generated, false otherwise.
		 */
		bool checkMetadata(OAMetadata &mm);
		
		// Resolution
		/*! In case of a need for resolution, wrap contract,
		 * counter, verifiable escrow of warranty, receipt and metadata;
		 * and send them to judge as an OAClaim object
		 * @return OAClaim object
		 */
		OAClaim* sendClaim() const;
		/*! Receive endorsement of warranty, when server is found guilty,
		 * check endorsement and return whether endorsement verifies
		 * @return true if endorsement verifies with the unendorsed coin of warranty, false otherwise
		 */
		bool checkEndorsement(vector<ZZ> &endorsement);

	protected:
		OAEscrowMessage* sendPaymentEscrow();
		void makeCoin(Wallet* wallet, const ZZ& R);
		void setCoin(const Coin& c);

	private:
		// set at construction
		const int stat;
		int counter;
		const VEPublicKey* verifiablePK;
		const VEPublicKey* regularPK;
		
		// set at setup phase
		ZZ id;
		Coin coin;
		OAContract* contract;
		OAEscrowMessage* warrantyVEMessage;
		
		// set at setup and replaced after every update
		OAMetadata* metadata; 
		
		Signature::Key* signKey; 		// created at construction
		Signature::Key* serversSignPK; 	// obtained from contract at setup phase
		string* receipt; 				// obtained upon payment
		vector<ZZ> endorsement; 		// obtained upon a valid claim
};

#endif

