#ifndef _OACONTRACT_H_
#define _OACONTRACT_H_

#include <time.h>
#include "Signature.h"

/*! \brief This class stores contract information 
 * on which interaction between client and server relies */

class OAContract {
	
	public:
		//! Constructor
		/*! OAContract stores due date, session ID,
		 * number of maximum updates allowed,
		 * and public keys of client and server */
		OAContract(const time_t t, const ZZ& id, const int maxUpdates, const Signature::Key &cSignPK, const Signature::Key &sSignPK)
			:due(t), id(id), maxUpdates(maxUpdates),
			clientsSignPK(new Signature::Key(cSignPK)),
			serversSignPK(new Signature::Key(sSignPK)) {}

		//! Copy Constructor
		OAContract(const OAContract &o)
			: due(o.due), id(o.id), maxUpdates(o.maxUpdates),
			clientsSignPK(new Signature::Key(*o.clientsSignPK)),
			serversSignPK(new Signature::Key(*o.serversSignPK)) {}
		
		//! Serialization Constructor
		OAContract(const string& str) { loadString(*this, str); }

		//! Destructor
		~OAContract() {	delete clientsSignPK;
						delete serversSignPK; }

		// getters
		/*! @return the session id of interaction between client and server */
		const ZZ& getID() const {return id;}
		/*! @return expiration time of contract */
		time_t getDue() const {return due;}
		/*! @return the number of maximum updates allowed */
		int getMaxUpdates() const {return maxUpdates;}
		/*! @return server's public key to verify its signatures */
		Signature::Key* getServersSignPK() const {return serversSignPK;}
		/*! @return client's public key to verify its signatures */
		Signature::Key* getClientsSignPK() const {return clientsSignPK;}

	private:
		time_t due;
		ZZ id;
		int maxUpdates;
		Signature::Key* clientsSignPK;
		Signature::Key* serversSignPK;

		//! Boost Serializer for OAContract class
		friend class boost::serialization::access;
		template <class Archive> 
		void serialize(Archive& ar, const unsigned int ver) {
			ar  & auto_nvp(due)
				& auto_nvp(id)
				& auto_nvp(maxUpdates)
				& auto_nvp(*clientsSignPK)
				& auto_nvp(*serversSignPK);
		}
};

#endif //_OACONTRACT_H_

