#ifndef _OAEscrowMessage_H_
#define _OAEscrowMessage_H_

#include "Coin.h"
#include "VECiphertext.h"
#include "Signature.h"
#include "VEPublicKey.h"

/*! \brief This class is a wrapper for the payment and warranty messages
 * sent in Step5 & Step6 of the Official Arbitration protocol */

class OAEscrowMessage {

	public:
		//! Constructor
		/*! Constructor takes in unendorsed coin, verifiable escrow (verifiable 
		 * encryption of the endorsement for the coin), and the public key for
		 * the sender's signature scheme (label for the verifiable escrow) */
		OAEscrowMessage(const Coin &coinPrime, VECiphertext* const escrow, 
                	   const Signature::Key &signPK)
		: coinPrime(coinPrime), escrow(escrow), signPK(signPK.getPublicKey()) {}

		//! Copy Constructor
		OAEscrowMessage(const OAEscrowMessage &o)
			: coinPrime(o.coinPrime), escrow(o.escrow),
			  signPK(new Signature::Key(*o.signPK)) {}
		
		//! Destructor
        ~OAEscrowMessage() { delete signPK;
							 delete escrow; }

		OAEscrowMessage(const string& s, const BankParameters *params) {
			// need to set params for Coin contained in message
			loadGZString(make_nvp("OAEscrowMessage", *this), s); 
			coinPrime.setParameters(params);
		}
		
		/*! Verify unendorsed coin and escrow */
		bool check(const VEPublicKey* pk, const int stat, const ZZ& R) const;
		
		// getters for coin', escrow, and signature PK
		/*! @return unendorsed coin*/
		Coin getCoinPrime() const { return coinPrime; }
		/*! @return verifiable escrow of endorsement*/
		VECiphertext* getEscrow() const { return escrow; }
		/*! @return public key of the coin owner (label of the verifiable escrow) */
		const Signature::Key* getPK() const { return signPK; }
				
	private:
		Coin coinPrime;
		VECiphertext* escrow;
		Signature::Key* signPK;

		//! Boost Serializer of OAEscrowMessage class
        friend class boost::serialization::access;
        template <class Archive> 
        void serialize(Archive& ar, const unsigned int ver) {
            ar  & auto_nvp(coinPrime)
                & auto_nvp(*escrow)
                & auto_nvp(*signPK);
        }
		
};

#endif /*OAEscrowMessage_H_*/

