#ifndef _OAJUDGE_H_
#define _OAJUDGE_H_

#include "VEDecrypter.h"
#include "OAClaim.h"
#include "OAUpdate.h"

//! \brief This class runs the judge(arbiter) side of the Official Arbitration protocol

class OAJudge {
	public:
		//! Constructor
		/**
		 * @param R 			randomization variable
		 * @param stat 			statistical security parameter
		 * @param verifiablePK	verifiable public key of arbiter
		 * @param regularpk 	regular public key of arbiter
		 * @param vD			verifiable escrow decrypter
		 * @param h				hashAlg to be used in endorsement devryption
		 */
		OAJudge(const ZZ &R, int stat, VEPublicKey* const verifiablePK, VEPublicKey* const regularPK,
				const VEDecrypter* vD, const hashalg_t &h)
					: R(R), stat(stat), verifiablePK(verifiablePK), regularPK(regularPK),
					verifiableDecrypter(vD), hashAlg(h) {}
		
		/*! Check the information sent in claim:
		 * Check contract date, receipt, and metadata  
		 * @param claim claim sent by client
		 * @return true if the information sent is valid, false otherwise
		 */
		bool checkClaim(const OAClaim &claim);
		
		/*! Check the signature on counter,
		 * and decide the dispute type by comparing counters
		 * @param counter_c counter that server claims to be client's
		 * @param signature client's signature on counter
		 * @return	OA_SERVER_PAYS if signature does not verify or server's counter (sent by client) is greater than counter_c, 
		 * OA_CLIENT_PAYS if counter_c is greater than server's counter +1,
		 * OA_DENIAL_OF_UPDATE if counter_c is equal to server's counter +1,
		 * OA_CORRUPTION if counter_c is equal to server's counter
		 */
		int compareCounters(int counter_c, const string &signature);

		// Cheater pays
		/*! Punish client 
		 */
		int clientPays() {return 0;}
		/*! Judge decrypts the escrow of warranty
		 * @return endorsement of warranty coin (NULL if the escrow does not verify)
		 */
		vector<ZZ>* serverPays();
		
		// In case of claim of denial of update
		/*! Check client's signature on counter, if it verifies
		 * impersonate the client and forward the update request 
		 * @param update update request sent by client
		 * @return update request of client (NULL if the signature does not verify)
		 */
		OAUpdate* update(OAUpdate* update);
		/*! Check the metadata message sent by server
		 * and forward it if it is valid
		 * @param mm metadata message constructed by server
		 * @return metadata message sent by server (NULL if it is not valid)
		 */
		OAMetadata* checkMetadata(OAMetadata* mm);
		
		// in case of claim of corruption
		/*! check the proof sent by server,
		 * return true if it is valid
		 * @param proof proof of integrity of the file, generated by server
		 * @return true if the proof is valid, false otherwise.
		 */
		bool checkProof(string &proof);
	
	private:
		// Following are constructor parameters
		ZZ R;
		int stat;
		VEPublicKey* verifiablePK;
		VEPublicKey* regularPK;
		const VEDecrypter* verifiableDecrypter;
		hashalg_t hashAlg;
		
		// Following are received with claim
		Signature::Key* clientsPK;
		Signature::Key* serversPK;
		Coin coinPrime;
		VECiphertext* escrow;
		OAContract* contract;
		int counter_s;
		
		// Asked for comparison
		int counter_c;
		
		// Asked in case of denial of update
		OAMetadata* metadata;
};

#endif

