#ifndef _OASERVER_CPP_
#define _OASERVER_CPP_

#include "VEProver.h"
#include "OA.h"
#include "OAServer.h"

//----------------------------------------------------------------------------//
// Constructors
OAServer::OAServer(const VEPublicKey* pk, const VEPublicKey* regularpk, 
					const int stat, const Signature::Key* sk)
			: stat(stat), counter(0), verifiablePK(pk), regularPK(regularpk),
			contract(NULL), paymentVEMessage(NULL), signKey(NULL),
			clientsSignPK(NULL), signature(NULL)
{
	signKey = new Signature::Key(*sk);
}

OAServer::OAServer(const OAServer &o)
	: stat(o.stat), counter(o.counter), verifiablePK(o.verifiablePK), regularPK(o.regularPK), 
	  id(o.id), coin(o.coin), contract(new OAContract(*o.contract)), 
	  paymentVEMessage(new OAEscrowMessage(*o.paymentVEMessage)), signKey(new Signature::Key(*o.signKey)), 
	  clientsSignPK(new Signature::Key(*o.clientsSignPK)), signature(new string(*o.signature)),
	  endorsement(o.endorsement), r(o.r) {}

//----------------------------------------------------------------------------//
// Destructor
OAServer::~OAServer() {
	reset();
	delete signKey;
}

void OAServer::reset() {
	if (contract!=NULL) delete contract;
	if (paymentVEMessage!=NULL) delete paymentVEMessage;
	if (clientsSignPK!=NULL) delete clientsSignPK;
	if (signature!=NULL) delete signature;
	counter=0;
	return;
}

//----------------------------------------------------------------------------//
// Setup
OAContract* OAServer::setup(Signature::Key& clientsPK) {
	// set up session id
	r = RandomBits_ZZ(2*stat);
	id = Hash::hash(r, verifiablePK->hashAlg, verifiablePK->hashKey);	
	
	// set client's publick key
	clientsSignPK = &clientsPK;
	
	// calculate contract due
	time_t due = time(NULL) + OA_CONTRACT_PERIOD;
	
	// create, store, and send contract
	Signature::Key* signPK = signKey->getPublicKey();
	contract = new OAContract(due, id, OA_MAX_UPDATES, clientsPK, *signPK);
	delete signPK;
	return contract;
}

OAMetadata* OAServer::sendMetadata(const string &clientsFile) {
	// delete the file (should be stored before, if it is to be used further!)
	delete &clientsFile;
	
	// create metadata and proof
	string metadata, proof, signature;
	metadata.resize(OA_METADATA_SIZE);
	proof.resize(OA_PROOF_SIZE);
	
	// sign the proof
	signature = Signature::sign(*signKey, saveString(counter)+(proof), regularPK->hashAlg);
	
	// send metadata message
	metadataMessage = new OAMetadata(metadata, proof, signature);
	return metadataMessage;
}

//----------------------------------------------------------------------------//
// Send Verifiable Escrow
OAEscrowMessage* OAServer::sendWarrantyEscrow(Wallet *wallet, const ZZ &R, OAEscrowMessage* paymentVEMessage) {
	// check escrow
	if (!paymentVEMessage->check(verifiablePK, stat, R))
		return NULL;
	this->paymentVEMessage = paymentVEMessage;

	makeCoin(wallet, R);
	return sendWarrantyEscrow();
}

OAEscrowMessage* OAServer::sendWarrantyEscrow() {
	// set up prover for forming escrow
	VEProver prover(verifiablePK);
	
	ZZ eCom = coin.getEndorsementCom();
	// the label needs to be the public key for the signature scheme
	// create the verifiable escrow
	VECiphertext* escrow = new VECiphertext(prover.verifiableEncrypt(eCom, 
											endorsement, coin.getCashGroup(),
											saveString(*contract),
											verifiablePK->hashAlg, stat));

	return new OAEscrowMessage(coin, escrow, *signKey);
}

void OAServer::makeCoin(Wallet* wallet, const ZZ& R) {
	// get a coin
	Coin coin = wallet->nextCoin(R);
	setCoin(coin);
}

void OAServer::setCoin(const Coin& c) {	
	// separate coin' from endorsement
	coin = c;
	endorsement = coin.getEndorsement();
	coin.unendorse();
}

//----------------------------------------------------------------------------//
// Get and send signatures on counters
string* OAServer::signCtrMt(string& clientSgnCtr) {
	// check signature
	if (!Signature::verify(*clientsSignPK, clientSgnCtr, saveString(counter), regularPK->hashAlg))
		return NULL;
	
	// store signature
	signature = &clientSgnCtr;
	
	// sign counter+metadata
	return new string (Signature::sign(*signKey, saveString(counter)+(*metadataMessage->getMetadata()), regularPK->hashAlg));
}

//----------------------------------------------------------------------------//
// Check payment and send receipt
bool OAServer::checkEndorsement(const vector<ZZ> &endorsement) {
	// check payment
	return paymentVEMessage->getCoinPrime().endorse(endorsement);
}

string* OAServer::sendReceipt() const {
	// send receipt
	return new string (Signature::sign(*signKey, saveString(*contract), regularPK->hashAlg));
}

//----------------------------------------------------------------------------//
// Accept updates
OAMetadata* OAServer::sendMetadata(OAUpdate* update) {
	// check signature on proof
	if (!Signature::verify(*clientsSignPK, *update->getSignOnCtr(), saveString(counter+1), regularPK->hashAlg))
		return NULL;
		
	// delete old signature, store new one.
	delete signature;
	signature = update->getSignOnCtr();
	delete update;
	
	// increment counter
	counter++;

	// create metadata and proof
	string metadata, proof, sign;
	metadata.resize(OA_METADATA_SIZE);
	proof.resize(OA_PROOF_SIZE);
	
	// sign proof
	sign = Signature::sign(*signKey, saveString(counter)+(metadata), regularPK->hashAlg);
	
	// create and send metadata
	metadataMessage = new OAMetadata(metadata, proof, sign);
	return metadataMessage;
}

// send proof (in case of claim of corruption)
string* OAServer::sendProof() const {
	string* proof = new string;
	proof->resize(OA_PROOF_SIZE);
	return proof;
}


#endif

