#ifndef _OASERVER_H_
#define _OASERVER_H_

#include "Wallet.h"
#include "OAContract.h"
#include "OAEscrowMessage.h"
#include "OAMetadata.h"
#include "OAUpdate.h"

//! \brief This class runs the server side of the Official Arbitration protocol

class OAServer {
	public:
		//! Constructor
		/**
		 * @param pk 		verifiable public key of arbiter
		 * @param regularpk regular public key of arbiter
		 * @param stat 		statistical security parameter
		 * @param sk		signature key of server
		 */
		OAServer(const VEPublicKey* pk, const VEPublicKey* repk, 
				 const int stat, const Signature::Key* sk);
		
		//! Copy Constructor
		OAServer(const OAServer &o);
		
		//! Destructor
		~OAServer();
		
		//! Reset data members to initilization values
		void reset();

		// getters
		/*! @return server's counter */
		int getCounter() const {return counter;}
		/*! @return client's signature on counter */
		string* getSignature() const {return signature;}

		// Setup functions
		/*! Receive client's public key at setup phase,
		 * return a contract including this and many other information
		 * @param clientsPK client's public key
		 * @return contract 
		 */
		OAContract* setup(Signature::Key &clientsPK);
		/*! Receive client's file, generate metadata and proof of
		 * metadata, sign on proof, return all three in a metadata message
		 * @param clientsFile client's file to be stored on server
		 * @return metadata message consisting of metadata, proof and signature 
		 */
		OAMetadata* sendMetadata(const string &clientsFile);
		
		// Payment functions
		/*! Receive escrow message consisting of unendorsed payment coin, verifiable escrow of 
		 * the coin, and the label of verifiable escrow; verify coin and verifiable escrow;
		 * if they verify, create a coin, then create a verifiable escrow for its endorsement and
		 * return them together with contract as an escrow message
		 * @param wallet 			server's wallet from which warranty coin is generated 
		 * @param R					randomization variable
		 * @param paymentVEMessage	escrow message consisting of unendorsed coin, verifiable escrow and public key of the server
		 * @return escrow message of warranty consisting of unendorsed payment coin, verifiable escrow of the coin, and the label of verifiable escrow (NULL, if payment escrow message does not verify)
		 */
		OAEscrowMessage* sendWarrantyEscrow(Wallet *wallet, const ZZ &R, OAEscrowMessage* paymentVEMessage);
		/*! Check client's signature on counter, if it verifies,
		 * return server's signature on concatanation of counter and metadata
		 * @param clientSgnCtr	clients signature on counter
		 * @return server's signature on concatanation of counter and metadata (NULL if client's signature does not verify)
		 */
		string* signCtrMt(string &clientSgnCtr);
		/*! Send receipt (signature on contract) 
		 * @return receipt (signature on contract)  
		 */
		string* sendReceipt() const;
		/*! Check payment endorsement sent by client
		 * return whether it verifies 
		 * @param payment endorsement sent by client 
		 * @return true if endorsement verifies, false otherwise
		 */
		bool checkEndorsement(const vector<ZZ> &endorsement);
		
		// Update rounds
		/*! Receive update information from client, check client's
		 * signature on counter, if it verifies create metadata and proof,
		 * sign on metadata and return them as an OAMetadata object 
		 * @param update update information
		 * @return metadata message consisting of metadata, proof of metadata and server's signature on metadata
		 */
		OAMetadata* sendMetadata(OAUpdate* update);
		
		// Resolve
		/*! Send proof of integrity to judge
		 * @return proof of integrity
		 */
		string* sendProof() const;

	protected:
		OAEscrowMessage* sendWarrantyEscrow();
		void makeCoin(Wallet* wallet, const ZZ& R);
		void setCoin(const Coin& c);
		
	private:
		// set at construction
		int stat;
		int counter;
		const VEPublicKey* verifiablePK;
		const VEPublicKey* regularPK;
		
		// set at setup phase
		ZZ id;
		Coin coin;
		OAContract* contract;
		OAEscrowMessage* paymentVEMessage;
		OAMetadata *metadataMessage;
		
		Signature::Key* signKey;		// set at construction
		Signature::Key* clientsSignPK;	// set at setup phase
		string* signature;				// signature on counter, set at setup and upon every update
		vector<ZZ> endorsement;			// obtained payment
		ZZ r;
};

#endif

